from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    AsyncIterator,
    Generic,
    Literal,
    Protocol,
    Sequence,
    TypeVar,
    Union,
)

from google.protobuf.json_format import MessageToDict
from typing_extensions import runtime_checkable

from .cache import FR24Cache
from .grpc import (
    BoundingBox,
    FlightDetailsParams,
    FollowFlightParams,
    LiveFeedParams,
    LiveFeedPlaybackParams,
    LiveFlightsStatusParams,
    NearestFlightsParams,
    PlaybackFlightParams,
    TopFlightsParams,
    flight_details,
    flight_details_df,
    follow_flight_stream,
    live_feed,
    live_feed_df,
    live_feed_playback,
    live_feed_playback_df,
    live_flights_status,
    live_flights_status_df,
    nearest_flights,
    nearest_flights_df,
    playback_flight,
    playback_flight_df,
    top_flights,
    top_flights_df,
)
from .json import (
    AirportListParams,
    FindParams,
    FlightListParams,
    PlaybackParams,
    airport_list,
    airport_list_parse,
    find,
    find_parse,
    flight_list,
    flight_list_df,
    flight_list_parse,
    playback,
    playback_df,
    playback_metadata_dict,
    playback_parse,
)
from .proto import SupportsToProto, parse_data
from .proto.v1_pb2 import (
    FlightDetailsResponse,
    FollowFlightResponse,
    LiveFeedResponse,
    LiveFlightsStatusResponse,
    NearestFlightsResponse,
    PlaybackFlightResponse,
    PlaybackResponse,
    RestrictionVisibility,
    TopFlightsResponse,
)
from .types import IntoFlightId, IntoTimestamp
from .types.cache import TabularFileFmt
from .types.grpc import LiveFeedField
from .types.isqx import (
    DistanceM,
    DurationS,
    LatitudeDeg,
    LongitudeDeg,
    TimestampS,
)
from .types.json import (
    FLIGHT_LIST_EMPTY,
    AirportList,
    Find,
    FlightList,
    Playback,
)
from .utils import (
    FileExistsBehaviour,
    FileLike,
    SupportsToDict,
    SupportsToPolars,
    dataclass_frozen,
    dataclass_opts,
    get_current_timestamp,
    parse_server_timestamp,
    static_check_signature,
    write_table,
)

if TYPE_CHECKING:
    import httpx
    import polars as pl
    from typing_extensions import TypeAlias

    from . import HTTPClient

logger = logging.getLogger(__name__)

#
# important traits and dataclasses
#


@dataclass_frozen
class ServiceFactory:
    http: HTTPClient

    def build_flight_list(self) -> FlightListService:
        return FlightListService(self)

    def build_playback(self) -> PlaybackService:
        return PlaybackService(self)

    def build_live_feed(self) -> LiveFeedService:
        return LiveFeedService(self)

    def build_live_feed_playback(self) -> LiveFeedPlaybackService:
        return LiveFeedPlaybackService(self)

    def build_airport_list(self) -> AirportListService:
        return AirportListService(self)

    def build_find(self) -> FindService:
        return FindService(self)

    def build_nearest_flights(self) -> NearestFlightsService:
        return NearestFlightsService(self)

    def build_live_flights_status(self) -> LiveFlightsStatusService:
        return LiveFlightsStatusService(self)

    def build_flight_details(self) -> FlightDetailsService:
        return FlightDetailsService(self)

    def build_top_flights(self) -> TopFlightsService:
        return TopFlightsService(self)

    def build_follow_flight(self) -> FollowFlightService:
        return FollowFlightService(self)

    def build_playback_flight(self) -> PlaybackFlightService:
        return PlaybackFlightService(self)


RequestT = TypeVar("RequestT")
"""Arguments for the request"""


@runtime_checkable
class SupportsFetch(Protocol[RequestT]):
    async def fetch(self, *args: Any, **kwargs: Any) -> APIResult[RequestT]:
        """Fetches data from the API."""
        ...


@dataclass_frozen
class APIResult(Generic[RequestT]):
    """Wraps the raw `Response` with request context.

    Note that at this stage, the response holds the *raw* bytes, possibly
    encoded with a scheme. Retrieve the raw bytes with `response.content` or
    parse it into json with `response.json()`.
    """

    request: RequestT
    response: httpx.Response


WriteLocation: TypeAlias = Union[FileLike, FR24Cache]


@runtime_checkable
class SupportsWriteTable(Protocol):
    def write_table(
        self,
        file: WriteLocation,
        *,
        format: TabularFileFmt,
        when_file_exists: FileExistsBehaviour = "backup",
    ) -> None:
        """Writes the object to the given file path."""


#
# definitions for services and results
#


@dataclass_frozen
class FlightListService(SupportsFetch[FlightListParams]):
    """Flight list service."""

    _factory: ServiceFactory

    @static_check_signature(FlightListParams)
    async def fetch(
        self,
        reg: str | None = None,
        flight: str | None = None,
        page: int = 1,
        limit: int = 10,
        timestamp: IntoTimestamp | Literal["now"] | None = "now",
        filter_by: Literal["historic"] | None = None,
    ) -> FlightListResult:
        """Fetch the flight list.

        :param reg: Aircraft registration (e.g. `B-HUJ`)
        :param flight: Flight number (e.g. `CX8747`)
        :param page: Page number
        :param limit: Number of results per page - use `100` if authenticated.
        :param timestamp: Show flights with ATD before this Unix timestamp
        :param filter_by: If `historic`, do not return scheduled or live flights
        """
        params = FlightListParams(
            reg=reg,
            flight=flight,
            page=page,
            limit=limit,
            timestamp=timestamp,
            filter_by=filter_by,
        )
        response = await flight_list(
            self._factory.http.client,
            params,
            self._factory.http.json_headers,
            self._factory.http.auth,
        )
        return FlightListResult(
            request=params,
            response=response,
        )

    @dataclass(**dataclass_opts)
    class FetchAllArgs(FlightListParams):
        """Arguments for fetching all pages of the flight list."""

        delay: DurationS[int] = field(default=5)
        """Delay between requests in seconds."""
        max_pages: int | None = field(default=None)
        """Maximum number of pages to fetch."""

    @static_check_signature(FetchAllArgs)
    async def fetch_all(
        self,
        reg: str | None = None,
        flight: str | None = None,
        page: int = 1,
        limit: int = 10,
        timestamp: IntoTimestamp | Literal["now"] | None = "now",
        filter_by: Literal["historic"] | None = None,
        delay: DurationS[int] = 5,
        max_pages: int | None = None,
    ) -> AsyncIterator[FlightListResult]:
        """Fetch all pages of the flight list.

        :param reg: Aircraft registration (e.g. `B-HUJ`)
        :param flight: Flight number (e.g. `CX8747`)
        :param page: Page number
        :param limit: Number of results per page - use `100` if authenticated.
        :param timestamp: Show flights with ATD before this Unix timestamp
        :param filter_by: If `historic`, do not return scheduled or live flights
        :param delay: Delay between requests in seconds.
        :param max_pages: Maximum number of pages to fetch.
        """
        # TODO: something nasty with async generators is happening here
        # (httpx leak)
        more = True
        current_timestamp = timestamp
        while more:
            result = await self.fetch(
                reg=reg,
                flight=flight,
                page=page,
                limit=limit,
                timestamp=current_timestamp,
                filter_by=filter_by,
            )
            response_dict = result.to_dict()
            # shouldn't happen, but stop in case we overshot
            if (data := response_dict["result"]["response"]["data"]) is None:
                break
            yield result
            if max_pages is not None and page >= max_pages:
                break
            page += 1

            # NOTE: for the next request, we have to both:
            # - update the timestamp to the earliest STD in the current batch
            # - increment the page
            # weird, but it's how the API works
            current_timestamp = min(
                t
                for d in data
                if (t := d["time"]["scheduled"]["departure"]) is not None
            )

            more = response_dict["result"]["response"]["page"]["more"]
            await asyncio.sleep(delay)

    def new_result_collection(self) -> FlightListResultCollection:
        """Create an empty list of flight list API results.

        Methods `to_dict` and `to_polars` can be used collect all unique rows in
        each flight list.
        """
        return FlightListResultCollection()


@dataclass_frozen
class FlightListResult(
    APIResult[FlightListParams],
    SupportsToDict[FlightList],
    SupportsToPolars,
    SupportsWriteTable,
):
    """A single result from the flight list API."""

    def to_dict(self) -> FlightList:
        return flight_list_parse(self.response).unwrap()

    def to_polars(self) -> pl.DataFrame:
        return flight_list_df(self.to_dict())

    def write_table(
        self,
        file: WriteLocation,
        *,
        format: TabularFileFmt = "parquet",
        when_file_exists: FileExistsBehaviour = "backup",
    ) -> None:
        if isinstance(file, FR24Cache):
            file = file.flight_list(self.request.kind).get_path(
                self.request.ident
            )
        write_table(
            self, file, format=format, when_file_exists=when_file_exists
        )


class FlightListResultCollection(
    list[FlightListResult],
    SupportsToDict[FlightList],
    SupportsToPolars,
    SupportsWriteTable,
):
    """A list of results from the flight list API."""

    def to_dict(self) -> FlightList:
        """Collects the raw bytes in each response into a single result.
        Duplicates are identified by their (flight id, time of departure),
        and are removed.

        No checking is made for the homogenity of the request parameters.
        """

        if len(self) == 0:
            return FLIGHT_LIST_EMPTY

        ident_hashes: set[int] = set()
        data = self[0].to_dict()
        flights_all = []
        for result in self:
            if (
                flights := result.to_dict()["result"]["response"]["data"]
            ) is None:
                continue
            for flight in flights:
                # NOTE: for future scheduled flights:
                # - the flight id is empty
                # - the standard time of departure is most certainly known
                # -> we use the STOD as the primary key for duplicates
                flight_id = flight["identification"]["id"]
                stod = flight["time"]["scheduled"]["departure"]
                ident_hash = hash((flight_id, stod))
                if ident_hash in ident_hashes:
                    if stod is not None:
                        logger.info(
                            f"skipping duplicate: {flight_id=} and {stod=}"
                        )
                        continue
                    else:
                        logger.warning(
                            f"unexpected empty STOD for {flight_id=}"
                        )
                ident_hashes.add(ident_hash)
                flights_all.append(flight)

        data["result"]["response"]["data"] = flights_all
        return data

    def to_polars(self) -> pl.DataFrame:
        return flight_list_df(self.to_dict())

    def write_table(
        self,
        file: WriteLocation,
        *,
        format: TabularFileFmt = "parquet",
        when_file_exists: FileExistsBehaviour = "backup",
    ) -> None:
        if not self:
            raise ValueError(
                "cannot save an empty flight list, "
                "must contain at least one valid flight list response"
            )
        if isinstance(file, FR24Cache):
            file = file.flight_list(self[0].request.kind).get_path(
                self[0].request.ident
            )
        write_table(
            self, file, format=format, when_file_exists=when_file_exists
        )


@dataclass_frozen
class PlaybackService(SupportsFetch[PlaybackParams]):
    """Playback service."""

    _factory: ServiceFactory

    @static_check_signature(PlaybackParams)
    async def fetch(
        self, flight_id: IntoFlightId, timestamp: IntoTimestamp | None = None
    ) -> PlaybackResult:
        """Fetch the playback data for a flight.

        :param flight_id: fr24 flight id, represented in hex
        :param timestamp: Actual time of departure (ATD) of the historic flight,
            Unix timestamp in seconds.
            Optional, but it is recommended to include it.
        """
        params = PlaybackParams(flight_id, timestamp)
        response = await playback(
            self._factory.http.client,
            params,
            self._factory.http.json_headers,
            self._factory.http.auth,
        )
        return PlaybackResult(
            request=params,
            response=response,
        )


@dataclass_frozen
class PlaybackResult(
    APIResult[PlaybackParams],
    SupportsToDict[Playback],
    SupportsToPolars,
    SupportsWriteTable,
):
    def to_dict(self) -> Playback:
        return playback_parse(self.response).unwrap()

    def to_polars(self) -> pl.DataFrame:
        return playback_df(self.to_dict())

    def metadata(self) -> dict[str, Any]:
        """Extracts flight metadata from the response."""
        return playback_metadata_dict(
            self.to_dict()["result"]["response"]["data"]["flight"]
        )

    def write_table(
        self,
        file: WriteLocation,
        *,
        format: TabularFileFmt = "parquet",
        when_file_exists: FileExistsBehaviour = "backup",
    ) -> None:
        if isinstance(file, FR24Cache):
            file = file.playback.get_path(self.request.flight_id)
        write_table(
            self, file, format=format, when_file_exists=when_file_exists
        )


@dataclass_frozen
class LiveFeedService(SupportsFetch[LiveFeedParams]):
    """Live feed service."""

    _factory: ServiceFactory

    @static_check_signature(LiveFeedParams)
    async def fetch(
        self,
        bounding_box: BoundingBox,
        stats: bool = False,
        limit: int = 1500,
        maxage: DurationS[int] = 14400,
        fields: set[LiveFeedField] = (
            lambda: {"flight", "reg", "route", "type"}
        )(),  # type: ignore
    ) -> LiveFeedResult:
        """Fetch the live feed.

        :param stats: Whether to include stats in the given area.
        :param limit: Maximum number of flights (should be set to 1500 for
            unauthorized users, 2000 for authorized users).
        :param maxage: Maximum time since last message update, seconds.
        :param fields: Fields to include. For unauthenticated users,
        a maximum of 4 fields can be included.
        When authenticated, `squawk`, `vspeed`, `airspace`, `logo_id` and `age`
        can be included.
        """
        params = LiveFeedParams(
            bounding_box=bounding_box,
            stats=stats,
            limit=limit,
            maxage=maxage,
            fields=fields,
        )
        response = await live_feed(
            self._factory.http.client,
            params.to_proto(),
            self._factory.http.grpc_headers,
        )
        # NOTE: serverTimeMs in the protobuf response would be more accurate
        timestamp = parse_server_timestamp(response) or get_current_timestamp()
        return LiveFeedResult(
            request=params,
            response=response,
            timestamp=timestamp,
        )


@dataclass_frozen
class LiveFeedResult(
    APIResult[LiveFeedParams],
    SupportsToProto[LiveFeedResponse],
    SupportsToDict[dict[str, Any]],
    SupportsToPolars,
    SupportsWriteTable,
):
    timestamp: TimestampS[int]

    def to_proto(self) -> LiveFeedResponse:
        return parse_data(self.response.content, LiveFeedResponse).unwrap()

    def to_dict(self) -> dict[str, Any]:
        return MessageToDict(self.to_proto(), preserving_proto_field_name=True)

    def to_polars(self) -> pl.DataFrame:
        return live_feed_df(self.to_proto())

    def write_table(
        self,
        file: WriteLocation,
        *,
        format: TabularFileFmt = "parquet",
        when_file_exists: FileExistsBehaviour = "backup",
    ) -> None:
        if isinstance(file, FR24Cache):
            file = file.live_feed.get_path(self.timestamp)
        write_table(
            self, file, format=format, when_file_exists=when_file_exists
        )


@dataclass_frozen
class LiveFeedPlaybackService(SupportsFetch[LiveFeedPlaybackParams]):
    """Live feed service."""

    _factory: ServiceFactory

    @static_check_signature(LiveFeedPlaybackParams)
    async def fetch(
        self,
        bounding_box: BoundingBox,
        stats: bool = False,
        limit: int = 1500,
        maxage: DurationS[int] = 14400,
        fields: set[LiveFeedField] = (
            lambda: {"flight", "reg", "route", "type"}
        )(),  # type: ignore
        timestamp: IntoTimestamp | Literal["now"] = "now",
        duration: DurationS[int] = 7,
        hfreq: int | None = None,
    ) -> LiveFeedPlaybackResult:
        """Fetch a playback of the live feed.

        :param stats: Whether to include stats in the given area.
        :param limit: Maximum number of flights (should be set to 1500 for
            unauthorized users, 2000 for authorized users).
        :param maxage: Maximum time since last message update, seconds.
        :param fields: Fields to include. For unauthenticated users, a maximum
        of 4 fields can be included.
        When authenticated, `squawk`, `vspeed`, `airspace`, `logo_id` and `age`
        can be included.
        :param timestamp: Start timestamp
        :param duration: Duration of prefetch, `floor(7.5*(multiplier))` seconds

        For 1x playback, this should be 7 seconds.
        :param hfreq: High frequency mode
        """
        params = LiveFeedPlaybackParams(
            bounding_box=bounding_box,
            stats=stats,
            limit=limit,
            maxage=maxage,
            fields=fields,
            timestamp=timestamp,
            duration=duration,
            hfreq=hfreq,
        )
        response = await live_feed_playback(
            self._factory.http.client,
            params.to_proto(),
            self._factory.http.grpc_headers,
        )
        return LiveFeedPlaybackResult(
            request=params,
            response=response,
        )


@dataclass_frozen
class LiveFeedPlaybackResult(
    APIResult[LiveFeedPlaybackParams],
    SupportsToProto[PlaybackResponse],
    SupportsToDict[dict[str, Any]],
    SupportsToPolars,
    SupportsWriteTable,
):
    def to_proto(self) -> PlaybackResponse:
        return parse_data(self.response.content, PlaybackResponse).unwrap()

    def to_dict(self) -> dict[str, Any]:
        return MessageToDict(self.to_proto(), preserving_proto_field_name=True)

    def to_polars(self) -> pl.DataFrame:
        return live_feed_playback_df(self.to_proto())

    def write_table(
        self,
        file: WriteLocation,
        *,
        format: TabularFileFmt = "parquet",
        when_file_exists: FileExistsBehaviour = "backup",
    ) -> None:
        if isinstance(file, FR24Cache):
            file = file.live_feed.get_path(self.request.timestamp)
        write_table(
            self, file, format=format, when_file_exists=when_file_exists
        )


@dataclass_frozen
class AirportListService(SupportsFetch[AirportListParams]):
    """Airport list service."""

    _factory: ServiceFactory

    @static_check_signature(AirportListParams)
    async def fetch(
        self,
        airport: str,
        mode: Literal["arrivals", "departures", "ground"],
        page: int = 1,
        limit: int = 10,
        timestamp: IntoTimestamp | Literal["now"] | None = "now",
    ) -> AirportListResult:
        """Fetch the airport list.

        :param airport: IATA airport code (e.g. `HKG`)
        :param mode: arrivals, departures or on ground aircraft
        :param page: Page number
        :param limit: Number of results per page - use `100` if authenticated.
        :param timestamp: Show flights with STA before this timestamp
        """
        params = AirportListParams(
            airport=airport,
            mode=mode,
            page=page,
            limit=limit,
            timestamp=timestamp,
        )
        response = await airport_list(
            self._factory.http.client,
            params,
            self._factory.http.json_headers,
            self._factory.http.auth,
        )
        return AirportListResult(
            request=params,
            response=response,
        )


@dataclass_frozen
class AirportListResult(
    APIResult[AirportListParams],
    SupportsToDict[AirportList],
):
    def to_dict(self) -> AirportList:
        """Parse the response into a dictionary."""
        return airport_list_parse(self.response).unwrap()


@dataclass_frozen
class FindService(SupportsFetch[FindParams]):
    """Find service."""

    _factory: ServiceFactory

    @static_check_signature(FindParams)
    async def fetch(self, query: str, limit: int = 50) -> FindResult:
        """Fetch the find results.

        :param query: Airport, schedule (HKG-CDG), or aircraft.
        """
        params = FindParams(query=query, limit=limit)
        response = await find(
            self._factory.http.client,
            params,
            self._factory.http.json_headers,
            self._factory.http.auth,
        )
        return FindResult(
            request=params,
            response=response,
        )


@dataclass_frozen
class FindResult(
    APIResult[FindParams],
    SupportsToDict[Find],
):
    """A single result from the find API."""

    def to_dict(self) -> Find:
        """Parse the response into a dictionary."""
        return find_parse(self.response).unwrap()


@dataclass_frozen
class NearestFlightsService(SupportsFetch[NearestFlightsParams]):
    """Nearest flights service."""

    _factory: ServiceFactory

    @static_check_signature(NearestFlightsParams)
    async def fetch(
        self,
        lat: LatitudeDeg[float],
        lon: LongitudeDeg[float],
        radius: DistanceM[int] = 10000,
        limit: int = 1500,
    ) -> NearestFlightsResult:
        """Fetch the nearest flights.

        :param lat: Latitude, degrees, -90 to 90
        :param lon: Longitude, degrees, -180 to 180
        :param radius: Radius, metres
        :param limit: Maximum number of aircraft to return
        """
        request = NearestFlightsParams(
            lat=lat,
            lon=lon,
            radius=radius,
            limit=limit,
        )
        response = await nearest_flights(
            self._factory.http.client,
            request.to_proto(),
            self._factory.http.grpc_headers,
        )
        timestamp = parse_server_timestamp(response) or get_current_timestamp()
        return NearestFlightsResult(
            request=request,
            response=response,
            timestamp=timestamp,
        )


@dataclass_frozen
class NearestFlightsResult(
    APIResult[NearestFlightsParams],
    SupportsToProto[NearestFlightsResponse],
    SupportsToDict[dict[str, Any]],
    SupportsToPolars,
    SupportsWriteTable,
):
    timestamp: TimestampS[int]

    def to_proto(self) -> NearestFlightsResponse:
        return parse_data(
            self.response.content, NearestFlightsResponse
        ).unwrap()

    def to_dict(self) -> dict[str, Any]:
        return MessageToDict(self.to_proto(), preserving_proto_field_name=True)

    def to_polars(self) -> pl.DataFrame:
        return nearest_flights_df(self.to_proto())

    def write_table(
        self,
        file: WriteLocation,
        *,
        format: TabularFileFmt = "parquet",
        when_file_exists: FileExistsBehaviour = "backup",
    ) -> None:
        if isinstance(file, FR24Cache):
            file = file.nearest_flights.get_path(
                self.request.lon, self.request.lat, self.timestamp
            )
        write_table(
            self, file, format=format, when_file_exists=when_file_exists
        )


@dataclass_frozen
class LiveFlightsStatusService(SupportsFetch[LiveFlightsStatusParams]):
    """Live flights status service."""

    _factory: ServiceFactory

    @static_check_signature(LiveFlightsStatusParams)
    async def fetch(
        self, flight_ids: Sequence[IntoFlightId]
    ) -> LiveFlightsStatusResult:
        """Fetch the live flights status.

        :param flight_ids: List of flight IDs to get status for
        """
        request = LiveFlightsStatusParams(flight_ids=flight_ids)
        response = await live_flights_status(
            self._factory.http.client,
            request.to_proto(),
            self._factory.http.grpc_headers,
        )
        timestamp = parse_server_timestamp(response) or get_current_timestamp()
        return LiveFlightsStatusResult(
            request=request,
            response=response,
            timestamp=timestamp,
        )


@dataclass_frozen
class LiveFlightsStatusResult(
    APIResult[LiveFlightsStatusParams],
    SupportsToProto[LiveFlightsStatusResponse],
    SupportsToDict[dict[str, Any]],
    SupportsToPolars,
    SupportsWriteTable,
):
    timestamp: TimestampS[int]

    def to_proto(self) -> LiveFlightsStatusResponse:
        return parse_data(
            self.response.content, LiveFlightsStatusResponse
        ).unwrap()

    def to_dict(self) -> dict[str, Any]:
        return MessageToDict(self.to_proto(), preserving_proto_field_name=True)

    def to_polars(self) -> pl.DataFrame:
        return live_flights_status_df(self.to_proto())

    def write_table(
        self,
        file: WriteLocation,
        *,
        format: TabularFileFmt = "parquet",
        when_file_exists: FileExistsBehaviour = "backup",
    ) -> None:
        if isinstance(file, FR24Cache):
            file = file.live_flights_status.get_path(self.timestamp)
        write_table(
            self, file, format=format, when_file_exists=when_file_exists
        )


@dataclass_frozen
class FollowFlightService:
    """Follow flight service for real-time streaming."""

    _factory: ServiceFactory

    @static_check_signature(FollowFlightParams)
    async def stream(
        self,
        flight_id: IntoFlightId,
        restriction_mode: (
            RestrictionVisibility.ValueType | str | bytes
        ) = RestrictionVisibility.NOT_VISIBLE,
    ) -> AsyncGenerator[FollowFlightResult, None]:
        """Stream real-time flight updates.

        :param flight_id: Flight ID to fetch details for.
        Must be live, or the response will contain an empty `DATA` frame error.
        :param restriction_mode: [FAA LADD](https://www.faa.gov/pilots/ladd)
            visibility mode.
        """
        request = FollowFlightParams(
            flight_id=flight_id, restriction_mode=restriction_mode
        )
        async for response in follow_flight_stream(
            self._factory.http.client,
            request.to_proto(),
            self._factory.http.grpc_headers,
        ):
            yield FollowFlightResult(
                request=request,
                response=response,
            )


@dataclass_frozen
class FollowFlightResult(
    SupportsToProto[FollowFlightResponse],
    SupportsToDict[dict[str, Any]],
):
    request: FollowFlightParams
    response: bytes

    def to_proto(self) -> FollowFlightResponse:
        return parse_data(self.response, FollowFlightResponse).unwrap()

    def to_dict(self) -> dict[str, Any]:
        return MessageToDict(self.to_proto(), preserving_proto_field_name=True)


@dataclass_frozen
class TopFlightsService(SupportsFetch[TopFlightsParams]):
    """Top flights service."""

    _factory: ServiceFactory

    @static_check_signature(TopFlightsParams)
    async def fetch(self, limit: int = 10) -> TopFlightsResult:
        """Fetch the top flights.

        :param limit: Maximum number of top flights to return (1-10)
        """
        request = TopFlightsParams(limit=limit)
        response = await top_flights(
            self._factory.http.client,
            request.to_proto(),
            self._factory.http.grpc_headers,
        )
        timestamp = parse_server_timestamp(response) or get_current_timestamp()
        return TopFlightsResult(
            request=request,
            response=response,
            timestamp=timestamp,
        )


@dataclass_frozen
class TopFlightsResult(
    APIResult[TopFlightsParams],
    SupportsToProto[TopFlightsResponse],
    SupportsToDict[dict[str, Any]],
    SupportsToPolars,
    SupportsWriteTable,
):
    timestamp: TimestampS[int]

    def to_proto(self) -> TopFlightsResponse:
        return parse_data(self.response.content, TopFlightsResponse).unwrap()

    def to_dict(self) -> dict[str, Any]:
        return MessageToDict(self.to_proto(), preserving_proto_field_name=True)

    def to_polars(self) -> pl.DataFrame:
        return top_flights_df(self.to_proto())

    def write_table(
        self,
        file: WriteLocation,
        *,
        format: TabularFileFmt = "parquet",
        when_file_exists: FileExistsBehaviour = "backup",
    ) -> None:
        if isinstance(file, FR24Cache):
            file = file.top_flights.get_path(self.timestamp)
        write_table(
            self, file, format=format, when_file_exists=when_file_exists
        )


@dataclass_frozen
class FlightDetailsService(SupportsFetch[FlightDetailsParams]):
    """Flight details service."""

    _factory: ServiceFactory

    @static_check_signature(FlightDetailsParams)
    async def fetch(
        self,
        flight_id: IntoFlightId,
        restriction_mode: (
            RestrictionVisibility.ValueType | str | bytes
        ) = RestrictionVisibility.NOT_VISIBLE,
        verbose: bool = True,
    ) -> FlightDetailsResult:
        """Fetch flight details.

        :param flight_id: Flight ID to fetch details for. Must be live, or the
            response will contain an empty `DATA` frame error.
        :param restriction_mode: [FAA LADD](https://www.faa.gov/pilots/ladd)
            visibility mode.
        :param verbose: Whether to include
            [fr24.proto.v1_pb2.FlightDetailsResponse.flight_plan]
            and [fr24.proto.v1_pb2.FlightDetailsResponse.aircraft_details] in
            the response.
        """
        request = FlightDetailsParams(
            flight_id=flight_id,
            restriction_mode=restriction_mode,
            verbose=verbose,
        )
        response = await flight_details(
            self._factory.http.client,
            request.to_proto(),
            self._factory.http.grpc_headers,
        )
        timestamp = parse_server_timestamp(response) or get_current_timestamp()
        return FlightDetailsResult(
            request=request,
            response=response,
            timestamp=timestamp,
        )


@dataclass_frozen
class FlightDetailsResult(
    APIResult[FlightDetailsParams],
    SupportsToProto[FlightDetailsResponse],
    SupportsToDict[dict[str, Any]],
    SupportsToPolars,
    SupportsWriteTable,
):
    timestamp: TimestampS[int]

    def to_proto(self) -> FlightDetailsResponse:
        return parse_data(self.response.content, FlightDetailsResponse).unwrap()

    def to_dict(self) -> dict[str, Any]:
        return MessageToDict(self.to_proto(), preserving_proto_field_name=True)

    def to_polars(self) -> pl.DataFrame:
        return flight_details_df(self.to_proto())

    def write_table(
        self,
        file: WriteLocation,
        *,
        format: TabularFileFmt = "parquet",
        when_file_exists: FileExistsBehaviour = "backup",
    ) -> None:
        if isinstance(file, FR24Cache):
            file = file.flight_details.get_path(
                self.request.flight_id, self.timestamp
            )
        write_table(
            self, file, format=format, when_file_exists=when_file_exists
        )


@dataclass_frozen
class PlaybackFlightService(SupportsFetch[PlaybackFlightParams]):
    """Playback flight service."""

    _factory: ServiceFactory

    @static_check_signature(PlaybackFlightParams)
    async def fetch(
        self, flight_id: IntoFlightId, timestamp: IntoTimestamp
    ) -> PlaybackFlightResult:
        """Fetch playback flight details.

        :param flight_id: Flight ID to fetch details for. Must not be live, or
            the response will contain an empty `DATA` frame error.
        :param timestamp: Actual time of departure (ATD) of the historic flight
        """
        request = PlaybackFlightParams(
            flight_id=flight_id,
            timestamp=timestamp,
        )
        response = await playback_flight(
            self._factory.http.client,
            request.to_proto(),
            self._factory.http.grpc_headers,
        )
        return PlaybackFlightResult(
            request=request,
            response=response,
        )


@dataclass_frozen
class PlaybackFlightResult(
    APIResult[PlaybackFlightParams],
    SupportsToProto[PlaybackFlightResponse],
    SupportsToDict[dict[str, Any]],
    SupportsToPolars,
    SupportsWriteTable,
):
    def to_proto(self) -> PlaybackFlightResponse:
        return parse_data(
            self.response.content, PlaybackFlightResponse
        ).unwrap()

    def to_dict(self) -> dict[str, Any]:
        return MessageToDict(self.to_proto(), preserving_proto_field_name=True)

    def to_polars(self) -> pl.DataFrame:
        return playback_flight_df(self.to_proto())

    def write_table(
        self,
        file: WriteLocation,
        *,
        format: TabularFileFmt = "parquet",
        when_file_exists: FileExistsBehaviour = "backup",
    ) -> None:
        if isinstance(file, FR24Cache):
            file = file.playback_flight.get_path(
                self.request.flight_id, self.request.timestamp
            )
        write_table(
            self, file, format=format, when_file_exists=when_file_exists
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: fr24/proto/_live_feed.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'fr24/proto/_live_feed.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from fr24.proto import _common_pb2 as fr24_dot_proto_dot___common__pb2
from google.protobuf import field_mask_pb2 as google_dot_protobuf_dot_field__mask__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x66r24/proto/_live_feed.proto\x1a\x18\x66r24/proto/_common.proto\x1a google/protobuf/field_mask.proto\"N\n\x12LocationBoundaries\x12\r\n\x05north\x18\x01 \x01(\x02\x12\r\n\x05south\x18\x02 \x01(\x02\x12\x0c\n\x04west\x18\x03 \x01(\x02\x12\x0c\n\x04\x65\x61st\x18\x04 \x01(\x02\"\xc6\x01\n\x12VisibilitySettings\x12)\n\x0csources_list\x18\x01 \x03(\x0e\x32\x13._common.DataSource\x12\'\n\rservices_list\x18\x02 \x03(\x0e\x32\x10._common.Service\x12*\n\x0ctraffic_type\x18\x03 \x01(\x0e\x32\x14._common.TrafficType\x12\x1c\n\x0fonly_restricted\x18\x04 \x01(\x08H\x00\x88\x01\x01\x42\x12\n\x10_only_restricted\"S\n\rAirportFilter\x12\x0c\n\x04iata\x18\x01 \x01(\t\x12\x12\n\ncountry_id\x18\x02 \x01(\x05\x12 \n\x04type\x18\x03 \x01(\x0e\x32\x12.AirportFilterType\"$\n\x08Interval\x12\x0b\n\x03min\x18\x01 \x01(\x05\x12\x0b\n\x03max\x18\x02 \x01(\x05\"?\n\rAirlineFilter\x12\x0c\n\x04icao\x18\x01 \x01(\t\x12 \n\x04type\x18\x02 \x01(\x0e\x32\x12.AirlineFilterType\"\xe4\x03\n\x06\x46ilter\x12\'\n\x14\x61ltitude_ranges_list\x18\x01 \x03(\x0b\x32\t.Interval\x12$\n\x11speed_ranges_list\x18\x02 \x03(\x0b\x32\t.Interval\x12%\n\rairlines_list\x18\x03 \x03(\x0b\x32\x0e.AirlineFilter\x12\x16\n\x0e\x63\x61llsigns_list\x18\x04 \x03(\t\x12\x13\n\x0bradars_list\x18\x05 \x03(\t\x12\x11\n\tregs_list\x18\x06 \x03(\t\x12%\n\rairports_list\x18\x07 \x03(\x0b\x32\x0e.AirportFilter\x12\x14\n\x0c\x66lights_list\x18\x08 \x03(\t\x12\x12\n\ntypes_list\x18\t \x03(\t\x12)\n\x16\x62irth_year_ranges_list\x18\n \x03(\x0b\x32\t.Interval\x12\x14\n\x0csquawks_list\x18\x0b \x03(\r\x12$\n\x0corigins_list\x18\x0c \x03(\x0b\x32\x0e.AirportFilter\x12)\n\x11\x64\x65stinations_list\x18\r \x03(\x0b\x32\x0e.AirportFilter\x12)\n\x0f\x63\x61tegories_list\x18\x0e \x03(\x0e\x32\x10._common.Service\x12\x16\n\x0e\x61irspaces_list\x18\x0f \x03(\t\"\xab\x03\n\x0fLiveFeedRequest\x12#\n\x06\x62ounds\x18\x01 \x01(\x0b\x32\x13.LocationBoundaries\x12%\n\x08settings\x18\x02 \x01(\x0b\x32\x13.VisibilitySettings\x12\x1d\n\x0c\x66ilters_list\x18\x03 \x01(\x0b\x32\x07.Filter\x12\x13\n\x0b\x66leets_list\x18\x04 \x01(\t\x12\x16\n\x0ehighlight_mode\x18\x05 \x01(\x08\x12\x12\n\x05stats\x18\x06 \x01(\x08H\x00\x88\x01\x01\x12\x12\n\x05limit\x18\x07 \x01(\x05H\x01\x88\x01\x01\x12\x13\n\x06maxage\x18\x08 \x01(\x05H\x02\x88\x01\x01\x12=\n\x10restriction_mode\x18\t \x01(\x0e\x32\x1e._common.RestrictionVisibilityH\x03\x88\x01\x01\x12.\n\nfield_mask\x18\n \x01(\x0b\x32\x1a.google.protobuf.FieldMask\x12 \n\x18selected_flight_ids_list\x18\x0b \x03(\x07\x42\x08\n\x06_statsB\x08\n\x06_limitB\t\n\x07_maxageB\x13\n\x11_restriction_mode\"\x9f\x01\n\x10LiveFeedResponse\x12%\n\x0c\x66lights_list\x18\x01 \x03(\x0b\x32\x0f._common.Flight\x12\x1d\n\x05stats\x18\x02 \x01(\x0b\x32\x0e._common.Stats\x12-\n\x14selected_flight_list\x18\x03 \x03(\x0b\x32\x0f._common.Flight\x12\x16\n\x0eserver_time_ms\x18\x04 \x01(\x04\"\x81\x01\n\x0fPlaybackRequest\x12+\n\x11live_feed_request\x18\x01 \x01(\x0b\x32\x10.LiveFeedRequest\x12\x11\n\ttimestamp\x18\x02 \x01(\x05\x12\x10\n\x08prefetch\x18\x03 \x01(\x05\x12\x12\n\x05hfreq\x18\x04 \x01(\x05H\x00\x88\x01\x01\x42\x08\n\x06_hfreq\"A\n\x10PlaybackResponse\x12-\n\x12live_feed_response\x18\x01 \x01(\x0b\x32\x11.LiveFeedResponse*8\n\x11\x41irportFilterType\x12\x08\n\x04\x42OTH\x10\x00\x12\x0b\n\x07INBOUND\x10\x01\x12\x0c\n\x08OUTBOUND\x10\x02*4\n\x11\x41irlineFilterType\x12\x0e\n\nPAINTED_AS\x10\x00\x12\x0f\n\x0bOPERATED_BY\x10\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'fr24.proto._live_feed_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_AIRPORTFILTERTYPE']._serialized_start=1838
  _globals['_AIRPORTFILTERTYPE']._serialized_end=1894
  _globals['_AIRLINEFILTERTYPE']._serialized_start=1896
  _globals['_AIRLINEFILTERTYPE']._serialized_end=1948
  _globals['_LOCATIONBOUNDARIES']._serialized_start=91
  _globals['_LOCATIONBOUNDARIES']._serialized_end=169
  _globals['_VISIBILITYSETTINGS']._serialized_start=172
  _globals['_VISIBILITYSETTINGS']._serialized_end=370
  _globals['_AIRPORTFILTER']._serialized_start=372
  _globals['_AIRPORTFILTER']._serialized_end=455
  _globals['_INTERVAL']._serialized_start=457
  _globals['_INTERVAL']._serialized_end=493
  _globals['_AIRLINEFILTER']._serialized_start=495
  _globals['_AIRLINEFILTER']._serialized_end=558
  _globals['_FILTER']._serialized_start=561
  _globals['_FILTER']._serialized_end=1045
  _globals['_LIVEFEEDREQUEST']._serialized_start=1048
  _globals['_LIVEFEEDREQUEST']._serialized_end=1475
  _globals['_LIVEFEEDRESPONSE']._serialized_start=1478
  _globals['_LIVEFEEDRESPONSE']._serialized_end=1637
  _globals['_PLAYBACKREQUEST']._serialized_start=1640
  _globals['_PLAYBACKREQUEST']._serialized_end=1769
  _globals['_PLAYBACKRESPONSE']._serialized_start=1771
  _globals['_PLAYBACKRESPONSE']._serialized_end=1836
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: fr24/proto/_live_flight_status.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'fr24/proto/_live_flight_status.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from fr24.proto import _common_pb2 as fr24_dot_proto_dot___common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$fr24/proto/_live_flight_status.proto\x1a\x18\x66r24/proto/_common.proto\"a\n\x14LiveFlightStatusData\x12\x0b\n\x03lat\x18\x01 \x01(\x02\x12\x0b\n\x03lon\x18\x02 \x01(\x02\x12\x1f\n\x06status\x18\x03 \x01(\x0e\x32\x0f._common.Status\x12\x0e\n\x06squawk\x18\x04 \x01(\r\"3\n\x18LiveFlightsStatusRequest\x12\x17\n\x0f\x66light_ids_list\x18\x01 \x03(\x07\"J\n\x10LiveFlightStatus\x12\x11\n\tflight_id\x18\x01 \x01(\r\x12#\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x15.LiveFlightStatusData\"C\n\x19LiveFlightsStatusResponse\x12&\n\x0b\x66lights_map\x18\x01 \x03(\x0b\x32\x11.LiveFlightStatusb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'fr24.proto._live_flight_status_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_LIVEFLIGHTSTATUSDATA']._serialized_start=66
  _globals['_LIVEFLIGHTSTATUSDATA']._serialized_end=163
  _globals['_LIVEFLIGHTSSTATUSREQUEST']._serialized_start=165
  _globals['_LIVEFLIGHTSSTATUSREQUEST']._serialized_end=216
  _globals['_LIVEFLIGHTSTATUS']._serialized_start=218
  _globals['_LIVEFLIGHTSTATUS']._serialized_end=292
  _globals['_LIVEFLIGHTSSTATUSRESPONSE']._serialized_start=294
  _globals['_LIVEFLIGHTSSTATUSRESPONSE']._serialized_end=361
# @@protoc_insertion_point(module_scope)



from datetime import datetime
from typing import TypeVar, Union

from typing_extensions import TypeAlias

from .isqx import TimestampS

M = TypeVar("M")
"""Method"""


IntoTimestamp: TypeAlias = Union[TimestampS[int], datetime]
"""A type that can be converted to a timestamp (in seconds)."""


IntFlightId: TypeAlias = int
"""Flight ID as an integer."""
StrFlightIdHex: TypeAlias = str
"""Flight ID as a hexadecimal string."""
IntoFlightId: TypeAlias = Union[IntFlightId, StrFlightIdHex, bytes]
"""A type that can be converted to a flight ID."""



import pytest

from fr24 import FR24, FR24Cache
from fr24.service import LiveFlightsStatusResult, NearestFlightsResult


@pytest.fixture
async def live_flights_status_result(
    fr24: FR24,
    nearest_flights_result: NearestFlightsResult,
) -> LiveFlightsStatusResult:
    flight_ids = [
        f.flight.flightid
        for f in nearest_flights_result.to_proto().flights_list[:5]
    ]

    result = await fr24.live_flights_status.fetch(flight_ids=flight_ids)
    return result


@pytest.mark.anyio
async def test_live_flights_status_simple(
    live_flights_status_result: LiveFlightsStatusResult,
) -> None:
    num_flights_requested = len(live_flights_status_result.request.flight_ids)
    num_flights = len(live_flights_status_result.to_proto().flights_map)
    assert num_flights_requested == num_flights

    df = live_flights_status_result.to_polars()
    assert df.height == num_flights
    assert df.width == 5


@pytest.mark.anyio
async def test_live_flights_status_file_ops(
    live_flights_status_result: LiveFlightsStatusResult, cache: FR24Cache
) -> None:
    timestamp = live_flights_status_result.timestamp
    ident = f"{timestamp}"
    fp = cache.path / "live_flights_status" / f"{ident}.parquet"
    fp.parent.mkdir(parents=True, exist_ok=True)
    fp.unlink(missing_ok=True)

    live_flights_status_result.write_table(cache)
    assert fp.exists(), (
        f"{fp} not in {list(f.name for f in fp.parent.glob('*'))}"
    )

    df_local = cache.live_flights_status.scan_table(ident).collect()
    assert df_local.equals(live_flights_status_result.to_polars())

    fp.unlink()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from collections import abc as abc
from fr24.proto import _common_pb2 as _common_pb2
from google.protobuf import descriptor as descriptor
from google.protobuf import message as message
from google.protobuf.internal import containers as containers
import builtins as builtins
import sys
import typing as typing

if sys.version_info >= (3, 10):
    from typing import TypeAlias as _TypeAlias
else:
    from typing_extensions import TypeAlias as _TypeAlias

DESCRIPTOR: descriptor.FileDescriptor

@typing.final
class HistoricTrailRequest(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    FLIGHT_ID_FIELD_NUMBER: builtins.int
    flight_id: builtins.int
    def __init__(
        self,
        *,
        flight_id: builtins.int = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["flight_id", b"flight_id"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...

@typing.final
class HistoricTrailResponse(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    RADAR_RECORDS_LIST_FIELD_NUMBER: builtins.int
    @builtins.property
    def radar_records_list(self) -> containers.RepeatedCompositeFieldContainer[_common_pb2.RadarHistoryRecord]: ...
    def __init__(
        self,
        *,
        radar_records_list: abc.Iterable[_common_pb2.RadarHistoryRecord] | None = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["radar_records_list", b"radar_records_list"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: fr24/proto/_follow_flight.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'fr24/proto/_follow_flight.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from fr24.proto import _common_pb2 as fr24_dot_proto_dot___common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1f\x66r24/proto/_follow_flight.proto\x1a\x18\x66r24/proto/_common.proto\"b\n\x13\x46ollowFlightRequest\x12\x11\n\tflight_id\x18\x01 \x01(\x07\x12\x38\n\x10restriction_mode\x18\x02 \x01(\x0e\x32\x1e._common.RestrictionVisibility\"\xe2\x02\n\x14\x46ollowFlightResponse\x12,\n\raircraft_info\x18\x01 \x01(\x0b\x32\x15._common.AircraftInfo\x12(\n\x0b\x66light_plan\x18\x02 \x01(\x0b\x32\x13._common.FlightPlan\x12,\n\rschedule_info\x18\x03 \x01(\x0b\x32\x15._common.ScheduleInfo\x12\x30\n\x0f\x66light_progress\x18\x04 \x01(\x0b\x32\x17._common.FlightProgress\x12\x30\n\x0b\x66light_info\x18\x05 \x01(\x0b\x32\x1b._common.ExtendedFlightInfo\x12.\n\x11\x66light_trail_list\x18\x06 \x03(\x0b\x32\x13._common.TrailPoint\x12\x30\n\x0fposition_buffer\x18\x07 \x01(\x0b\x32\x17._common.PositionBufferb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'fr24.proto._follow_flight_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_FOLLOWFLIGHTREQUEST']._serialized_start=61
  _globals['_FOLLOWFLIGHTREQUEST']._serialized_end=159
  _globals['_FOLLOWFLIGHTRESPONSE']._serialized_start=162
  _globals['_FOLLOWFLIGHTRESPONSE']._serialized_end=516
# @@protoc_insertion_point(module_scope)



import pytest
from pydantic import BaseModel, ConfigDict, Discriminator
from typing_extensions import Annotated

from fr24 import FR24
from fr24.types.json import (
    Entry,
    Info,
    Stats,
    is_aircraft,
    is_airport,
    is_live,
    is_operator,
    is_schedule,
)
from fr24.utils import get_current_timestamp


# overwriting the original FindResult for now
# https://github.com/python/typing/issues/1467
class FindResult(BaseModel):
    model_config = ConfigDict(extra="forbid")

    results: list[Annotated[Entry, Discriminator("type")]]
    stats: Stats
    info: Info


@pytest.mark.anyio
async def test_find_airport(fr24: FR24) -> None:
    list_ = (await fr24.find.fetch("tou")).to_dict()
    assert list_ is not None
    assert list_["stats"]["count"]["airport"] >= 2

    found_francazal = False
    for result in list_["results"]:
        assert is_airport(result)
        if "LFBF" in result["label"]:
            found_francazal = True
    assert found_francazal

    FindResult.model_validate(list_)


@pytest.mark.anyio
async def test_find_aircraft(fr24: FR24) -> None:
    list_ = (await fr24.find.fetch("b-hp")).to_dict()
    assert list_ is not None

    found_bhpb = False
    for result in list_["results"]:
        if is_aircraft(result) and result["id"] == "B-HPB":
            found_bhpb = True

    assert found_bhpb

    FindResult.model_validate(list_)


@pytest.mark.anyio
async def test_find_operator(fr24: FR24) -> None:
    list_ = (await fr24.find.fetch("cat")).to_dict()
    assert list_ is not None

    found_cathay = False
    for result in list_["results"]:
        if is_operator(result) and "Cathay" in result["label"]:
            found_cathay = True
    assert found_cathay

    FindResult.model_validate(list_)


@pytest.mark.anyio
async def test_find_schedule_and_live(fr24: FR24) -> None:
    # avoid deadzone in the middle of the night
    utc_h = get_current_timestamp() % 86400 // 3600
    route = "cju-gmp" if utc_h < 8 else "fco-mad" if utc_h < 16 else "lax-sfo"
    list_ = (await fr24.find.fetch(route)).to_dict()
    assert list_ is not None

    schedule_count = 0
    live_count = 0
    for result in list_["results"]:
        if is_schedule(result):
            schedule_count += 1
        elif is_live(result):
            live_count += 1
    assert list_["stats"]["count"]["schedule"] == schedule_count
    assert schedule_count > 0
    assert list_["stats"]["count"]["live"] == live_count
    assert live_count > 0

    FindResult.model_validate(list_)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: fr24/proto/_fetch_search_index.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'fr24/proto/_fetch_search_index.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$fr24/proto/_fetch_search_index.proto\"\xc2\x01\n\x10\x46lightSearchData\x12\x0b\n\x03reg\x18\x01 \x01(\t\x12\x11\n\tschd_from\x18\x02 \x01(\t\x12\x0f\n\x07schd_to\x18\x03 \x01(\t\x12\x0e\n\x06\x66light\x18\x04 \x01(\t\x12\x10\n\x08operator\x18\x05 \x01(\t\x12\x0f\n\x07\x61\x63_type\x18\x06 \x01(\t\x12\x0b\n\x03lat\x18\x07 \x01(\x02\x12\x0b\n\x03lon\x18\x08 \x01(\x02\x12\x10\n\x08\x63\x61llsign\x18\t \x01(\t\x12\n\n\x02id\x18\n \x01(\r\x12\x12\n\nrestricted\x18\x0b \x01(\x08\"\x19\n\x17\x46\x65tchSearchIndexRequest\"C\n\x18\x46\x65tchSearchIndexResponse\x12\'\n\x0c\x66lights_list\x18\x01 \x03(\x0b\x32\x11.FlightSearchDatab\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'fr24.proto._fetch_search_index_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_FLIGHTSEARCHDATA']._serialized_start=41
  _globals['_FLIGHTSEARCHDATA']._serialized_end=235
  _globals['_FETCHSEARCHINDEXREQUEST']._serialized_start=237
  _globals['_FETCHSEARCHINDEXREQUEST']._serialized_end=262
  _globals['_FETCHSEARCHINDEXRESPONSE']._serialized_start=264
  _globals['_FETCHSEARCHINDEXRESPONSE']._serialized_end=331
# @@protoc_insertion_point(module_scope)



# ruff: noqa
# mypy: disable-error-code="top-level-await, no-redef"
# %%
# --8<-- [start:script0]
import httpx
from fr24.grpc import nearest_flights
from fr24.proto.v1_pb2 import (
    NearestFlightsResponse,
    NearestFlightsRequest,
    Geolocation,
)
from fr24.proto import parse_data
from fr24.proto.headers import get_grpc_headers


async def nearest_flights_data() -> NearestFlightsResponse:
    headers = httpx.Headers(get_grpc_headers(auth=None))
    async with httpx.AsyncClient() as client:
        message = NearestFlightsRequest(
            location=Geolocation(lat=22.31257, lon=113.92708),
            radius=1000,
            limit=1500,
        )
        response = await nearest_flights(client, message, headers)
        return parse_data(response.content, NearestFlightsResponse).unwrap()


data = await nearest_flights_data()
data
# --8<-- [end:script0]
# %%
"""
# --8<-- [start:output0]
flights_list {
  flight {
    flightid: 905701578
    lat: 22.3095245
    lon: 113.930717
    track: 78
    speed: 9
    timestamp: 1720071634
    on_ground: true
    callsign: "CES502"
    extra_info {
      flight: "MU502"
      reg: "B-1908"
      route {
        from: "HKG"
        to: "PVG"
      }
      type: "B738"
      logo_id: 131
    }
  }
  distance: 504
}
flights_list {
  flight {
...
      logo_id: 335
    }
  }
  distance: 783
}
# --8<-- [end:output0]
"""



# ruff: noqa
# mypy: disable-error-code="top-level-await, no-redef"
# %%
# --8<-- [start:script0]
import httpx
from fr24.grpc import live_trail
from fr24.proto.v1_pb2 import LiveTrailRequest, LiveTrailResponse
from fr24.proto import parse_data
from fr24.proto.headers import get_grpc_headers


async def live_trail_data() -> LiveTrailResponse:
    headers = httpx.Headers(get_grpc_headers(auth=None))
    async with httpx.AsyncClient() as client:
        message = LiveTrailRequest(flight_id=0x395C43CF)
        response = await live_trail(client, message, headers)
        return parse_data(response.content, LiveTrailResponse).unwrap()


data = await live_trail_data()
data
# --8<-- [end:script0]
# %%
"""
# --8<-- [start:output0]
radar_records_list {
  timestamp: 1720064455
  lat: 30.2406578
  lon: 120.420403
  spd: 15
  heading: 61
  callsign: "CPA959"
}
radar_records_list {
  timestamp: 1720064462
  lat: 30.2408924
  lon: 120.420937
  heading: 61
  callsign: "CPA959"
}
radar_records_list {
  timestamp: 1720064480
  lat: 30.2415791
  lon: 120.422478
  spd: 18
  heading: 61
  callsign: "CPA959"
}
radar_records_list {
  timestamp: 1720064488
...
  heading: 269
  vspd: -960
  squawk: 13874
  callsign: "CPA959"
}
...
# --8<-- [end:output0]
"""



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from collections import abc as abc
from google.protobuf import descriptor as descriptor
from google.protobuf import message as message
from google.protobuf.internal import containers as containers
import builtins as builtins
import sys
import typing as typing

if sys.version_info >= (3, 10):
    from typing import TypeAlias as _TypeAlias
else:
    from typing_extensions import TypeAlias as _TypeAlias

DESCRIPTOR: descriptor.FileDescriptor

@typing.final
class FollowedFlight(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    FLIGHT_ID_FIELD_NUMBER: builtins.int
    LIVE_CLICKS_FIELD_NUMBER: builtins.int
    TOTAL_CLICKS_FIELD_NUMBER: builtins.int
    FLIGHT_NUMBER_FIELD_NUMBER: builtins.int
    CALLSIGN_FIELD_NUMBER: builtins.int
    SQUAWK_FIELD_NUMBER: builtins.int
    FROM_IATA_FIELD_NUMBER: builtins.int
    FROM_CITY_FIELD_NUMBER: builtins.int
    TO_IATA_FIELD_NUMBER: builtins.int
    TO_CITY_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    FULL_DESCRIPTION_FIELD_NUMBER: builtins.int
    flight_id: builtins.int
    """FR24 Flight ID (e.g. `962709904` = `0x3961c990`)"""
    live_clicks: builtins.int
    """Number of live clicks/views (e.g. `1212`)"""
    total_clicks: builtins.int
    """Total number of clicks/views (e.g. `6128`)"""
    flight_number: builtins.str
    """Flight number (e.g. `"EK261"`)"""
    callsign: builtins.str
    """Callsign (e.g. `"UAE261"`)"""
    squawk: builtins.int
    """Squawk code (e.g. `4032` = `0o7700`)"""
    from_iata: builtins.str
    """IATA code of origin airport (e.g. `"DXB"`)"""
    from_city: builtins.str
    """City name of origin airport (e.g. `"Dubai"`)"""
    to_iata: builtins.str
    """IATA code of destination airport (e.g. `"GRU"`)"""
    to_city: builtins.str
    """City name of destination airport (e.g. `"Sao Paulo"`)"""
    type: builtins.str
    """Aircraft type code (e.g. `"A388"`)"""
    full_description: builtins.str
    """Full aircraft description (e.g. `"Airbus A380-861"`)"""
    def __init__(
        self,
        *,
        flight_id: builtins.int = ...,
        live_clicks: builtins.int = ...,
        total_clicks: builtins.int = ...,
        flight_number: builtins.str = ...,
        callsign: builtins.str = ...,
        squawk: builtins.int = ...,
        from_iata: builtins.str = ...,
        from_city: builtins.str = ...,
        to_iata: builtins.str = ...,
        to_city: builtins.str = ...,
        type: builtins.str = ...,
        full_description: builtins.str = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["callsign", b"callsign", "flight_id", b"flight_id", "flight_number", b"flight_number", "from_city", b"from_city", "from_iata", b"from_iata", "full_description", b"full_description", "live_clicks", b"live_clicks", "squawk", b"squawk", "to_city", b"to_city", "to_iata", b"to_iata", "total_clicks", b"total_clicks", "type", b"type"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...

@typing.final
class TopFlightsRequest(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    LIMIT_FIELD_NUMBER: builtins.int
    limit: builtins.int
    """1..=10"""
    def __init__(
        self,
        *,
        limit: builtins.int = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["limit", b"limit"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...

@typing.final
class TopFlightsResponse(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    SCOREBOARD_LIST_FIELD_NUMBER: builtins.int
    @builtins.property
    def scoreboard_list(self) -> containers.RepeatedCompositeFieldContainer[FollowedFlight]: ...
    def __init__(
        self,
        *,
        scoreboard_list: abc.Iterable[FollowedFlight] | None = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["scoreboard_list", b"scoreboard_list"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...



# ruff: noqa
# mypy: disable-error-code="top-level-await, no-redef"
# %%
from fr24 import FR24
from fr24.types.json import AirportList

import polars as pl


async def my_arrivals() -> AirportList:
    async with FR24() as fr24:
        result = await fr24.airport_list.fetch(
            airport="tls",
            mode="arrivals",
        )
        return result.to_dict()


airports = await my_arrivals()
arrivals = airports["result"]["response"]["airport"]["pluginData"]["schedule"][
    "arrivals"
]["data"]
assert arrivals is not None
df = pl.json_normalize(arrivals)
print(df)
# %%
# --8<-- [start:script0]
import httpx

from fr24.types.json import AirportList
from fr24.json import airport_list, AirportListParams
from fr24.proto.headers import get_grpc_headers

import polars as pl


async def my_arrivals() -> AirportList:
    headers = httpx.Headers(get_grpc_headers(auth=None))
    async with httpx.AsyncClient() as client:
        response = await airport_list(
            client,
            AirportListParams(airport="tls", mode="arrivals"),
            headers=headers,
            auth=None,
        )
        response.raise_for_status()
        list_ = response.json()
        return list_  # type: ignore


airports = await my_arrivals()
arrivals = airports["result"]["response"]["airport"]["pluginData"]["schedule"][
    "arrivals"
]["data"]
assert arrivals is not None
df = pl.json_normalize(arrivals)
print(df)
# --8<-- [end:script0]
# %%
"""
# --8<-- [start:df0]
shape: (10, 76)
┌───────────┬───────────┬───────────┬───────────┬───┬───────────┬───────────┬───────────┬──────────┐
│ flight.id ┆ flight.id ┆ flight.id ┆ flight.id ┆ … ┆ flight.ti ┆ flight.ai ┆ flight.ow ┆ flight.a │
│ entificat ┆ entificat ┆ entificat ┆ entificat ┆   ┆ me.other. ┆ rcraft.im ┆ ner       ┆ irline   │
│ ion.id    ┆ ion.row   ┆ ion.numbe ┆ ion.numbe ┆   ┆ duration  ┆ ages      ┆ ---       ┆ ---      │
│ ---       ┆ ---       ┆ r.d…      ┆ r.a…      ┆   ┆ ---       ┆ ---       ┆ null      ┆ null     │
│ str       ┆ i64       ┆ ---       ┆ ---       ┆   ┆ null      ┆ null      ┆           ┆          │
│           ┆           ┆ str       ┆ str       ┆   ┆           ┆           ┆           ┆          │
╞═══════════╪═══════════╪═══════════╪═══════════╪═══╪═══════════╪═══════════╪═══════════╪══════════╡
│ 3963916d  ┆ 563063588 ┆ AF6132    ┆ null      ┆ … ┆ null      ┆ null      ┆ null      ┆ null     │
│           ┆ 1         ┆           ┆           ┆   ┆           ┆           ┆           ┆          │
│ null      ┆ 563365562 ┆ null      ┆ null      ┆ … ┆ null      ┆ null      ┆ null      ┆ null     │
│           ┆ 4         ┆           ┆           ┆   ┆           ┆           ┆           ┆          │
│ 39638d76  ┆ 563085809 ┆ T71527    ┆ null      ┆ … ┆ null      ┆ null      ┆ null      ┆ null     │
│           ┆ 2         ┆           ┆           ┆   ┆           ┆           ┆           ┆          │
│ 3963a803  ┆ 563063662 ┆ AF7408    ┆ null      ┆ … ┆ null      ┆ null      ┆ null      ┆ null     │
│           ┆ 2         ┆           ┆           ┆   ┆           ┆           ┆           ┆          │
│ 396399a4  ┆ 563087474 ┆ U24849    ┆ EC4849    ┆ … ┆ null      ┆ null      ┆ null      ┆ null     │
│           ┆ 7         ┆           ┆           ┆   ┆           ┆           ┆           ┆          │
│ null      ┆ 563088423 ┆ V72371    ┆ null      ┆ … ┆ null      ┆ null      ┆ null      ┆ null     │
│           ┆ 4         ┆           ┆           ┆   ┆           ┆           ┆           ┆          │
│ 3963978d  ┆ 563085809 ┆ T73718    ┆ null      ┆ … ┆ null      ┆ null      ┆ null      ┆ null     │
│           ┆ 1         ┆           ┆           ┆   ┆           ┆           ┆           ┆          │
│ 3963ab98  ┆ 563089434 ┆ XK720     ┆ null      ┆ … ┆ null      ┆ null      ┆ null      ┆ null     │
│           ┆ 8         ┆           ┆           ┆   ┆           ┆           ┆           ┆          │
│ 396398ab  ┆ 563071223 ┆ FR3903    ┆ null      ┆ … ┆ null      ┆ null      ┆ null      ┆ null     │
│           ┆ 6         ┆           ┆           ┆   ┆           ┆           ┆           ┆          │
│ 3963b61f  ┆ 563072081 ┆ GP155     ┆ null      ┆ … ┆ null      ┆ null      ┆ null      ┆ null     │
│           ┆ 9         ┆           ┆           ┆   ┆           ┆           ┆           ┆          │
└───────────┴───────────┴───────────┴───────────┴───┴───────────┴───────────┴───────────┴──────────┘
# --8<-- [end:df0]
"""



from __future__ import annotations

import secrets
from ..utils import DEFAULT_HEADERS

from ..types.json import Authentication

PLATFORM_VERSION = "25.061.0929"
# see ./README.md.


DEFAULT_HEADERS_GRPC = {
    **DEFAULT_HEADERS,
    "Accept": "*/*",
    "fr24-device-id": "web-000000000-000000000000000000000", # fingerprint
    "fr24-platform": f"web-{PLATFORM_VERSION}",
    "x-envoy-retry-grpc-on": "unavailable",
    "Content-Type": "application/grpc-web+proto",
    "X-User-Agent": "grpc-web-javascript/0.1",
    "X-Grpc-Web": "1",
    "DNT": "1",
}

def get_grpc_headers(*, auth: Authentication | None, device_id: None | str = None) -> dict[str, str]:
    headers = DEFAULT_HEADERS_GRPC.copy()
    if device_id is None:
        device_id = f"web-{secrets.token_urlsafe(32)}"
    headers["fr24-device-id"] = device_id
    if auth is not None and (token := auth["userData"].get("accessToken")) is not None:
        headers["authorization"] = f"Bearer {token}"
    return headers



from __future__ import annotations

import re

from textual import on
from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.widgets import Input, Label, Static

from fr24 import FR24
from fr24.types.json import is_aircraft, is_airport, is_schedule


def get_fr24(app: App[None]) -> FR24:
    return app.fr24  # type: ignore


class AirportWidget(Static):
    def compose(self) -> ComposeResult:
        assert self.name is not None
        self.iata = Static("", id="iata")
        self.icao = Static("", id="icao")
        self.fullname = Static("", id="name")
        self.airport_id = ""
        self.input = Input(id=self.name)

        yield Label(self.name)
        yield self.input
        with Horizontal():
            yield self.iata
            yield self.icao
        yield self.fullname

    @on(Input.Changed)
    def lookup_airport(self) -> None:
        self.app.query_one(AircraftWidget).input.value = ""
        self.app.query_one(FlightWidget).input.value = ""
        input_ = self.query_one(Input)
        if len(value := input_.value) >= 3:
            self.run_worker(
                self.update_airport(value),
                name="airport_lookup",
                exclusive=True,
            )
        else:
            self.update_info()
            self.airport_id = ""

    def update_info(self, info: None | dict[str, str] = None) -> None:
        # NOTE(abr): for some odd reason when i backspace too quickly it doesn't
        # clear properly
        if info is None:
            self.icao.update()
            self.iata.update()
            self.fullname.update()
            self.airport_id = ""
        else:
            self.icao.update(info["icao"])
            self.iata.update(info["iata"])
            self.fullname.update(info["name"])
            self.airport_id = info["iata"]

    async def update_airport(self, value: str) -> None:
        result = await get_fr24(self.app).find.fetch(query=value)
        if self.input.value != value:
            return  # stale worker, discard result.
        find_results = result.to_dict()
        if find_results is None:
            return self.update_info()
        candidate = next(
            (elt for elt in find_results["results"] if is_airport(elt)),
            None,
        )
        if candidate is None:
            return self.update_info()
        group = re.match(
            r"(?P<name>.+) \((?P<iata>\w+) / (?P<icao>\w+)\)",
            candidate["label"],
        )
        if group is None:
            return self.update_info()
        info: dict[str, str] = group.groupdict()
        return self.update_info(info)


class AircraftWidget(Static):
    def compose(self) -> ComposeResult:
        assert self.name is not None
        self.hex = Static("", id="hex")
        self.reg = Static("", id="reg")
        self.type = Static("", id="type")
        self.aircraft_id = ""
        self.input = Input(id=self.name)

        yield Label(self.name)
        yield self.input
        yield self.reg
        with Horizontal():
            yield self.hex
            yield self.type

    @on(Input.Changed)
    def lookup_aircraft(self) -> None:
        self.app.query_one(FlightWidget).input.value = ""
        for widget in self.app.query(AirportWidget):
            widget.input.value = ""
        input_ = self.query_one(Input)
        if len(value := input_.value) >= 3:
            self.run_worker(
                self.update_aircraft(value),
                name="aircraft_lookup",
                exclusive=True,
            )
        else:
            self.update_info()

    def update_info(
        self,
        reg: None | str = None,
        hex: None | str = None,
        typecode: None | str = None,
    ) -> None:
        if reg is None or hex is None or typecode is None:
            self.type.update()
            self.hex.update()
            self.reg.update()
            self.aircraft_id = ""
        else:
            self.type.update(typecode)
            self.hex.update(hex.lower())
            self.reg.update(reg)
            self.aircraft_id = reg

    async def update_aircraft(self, value: str) -> None:
        result = await get_fr24(self.app).find.fetch(query=value)
        if self.input.value != value:
            return
        res = result.to_dict()
        if res is None:
            return self.update_info()
        candidates = (elt for elt in res["results"] if is_aircraft(elt))
        aircraft = next(candidates, None)
        if aircraft is None:
            return self.update_info()

        return self.update_info(
            aircraft["id"],
            aircraft["detail"]["hex"],
            aircraft["detail"]["equip"],
        )


class FlightWidget(Static):
    def compose(self) -> ComposeResult:
        assert self.name is not None
        self.flight = Static("", id="flight")
        self.callsign = Static("", id="callsign")
        self.number = ""
        self.input = Input(id=self.name)

        yield Label(self.name)
        yield self.input
        with Horizontal():
            yield self.flight
            yield self.callsign

        self.input.focus()

    @on(Input.Changed)
    def lookup_number(self) -> None:
        self.app.query_one(AircraftWidget).input.value = ""
        for widget in self.app.query(AirportWidget):
            widget.input.value = ""
        input_ = self.query_one(Input)
        if len(value := input_.value) >= 3:
            self.run_worker(
                self.update_number(value), name="flight_lookup", exclusive=True
            )
        else:
            self.update_info()

    def update_info(
        self,
        number: str | None = None,
        callsign: str | None = None,
    ) -> None:
        if number is None or callsign is None:
            self.flight.update()
            self.callsign.update()
            self.number = ""
        else:
            self.flight.update(number)
            self.callsign.update(callsign)
            self.number = number

    async def update_number(self, value: str) -> None:
        result = await get_fr24(self.app).find.fetch(query=value)
        if self.input.value != value:
            return
        find_results = result.to_dict()
        if find_results is None:
            return self.update_info()
        candidate = next(
            (elt for elt in find_results["results"] if is_schedule(elt)),
            None,
        )
        if candidate is None:
            return self.update_info()
        return self.update_info(
            candidate["detail"]["flight"], candidate["detail"].get("callsign")
        )



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from collections import abc as abc
from google.protobuf import descriptor as descriptor
from google.protobuf import message as message
from google.protobuf.internal import containers as containers
import builtins as builtins
import sys
import typing as typing

if sys.version_info >= (3, 10):
    from typing import TypeAlias as _TypeAlias
else:
    from typing_extensions import TypeAlias as _TypeAlias

DESCRIPTOR: descriptor.FileDescriptor

@typing.final
class FlightSearchData(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    REG_FIELD_NUMBER: builtins.int
    SCHD_FROM_FIELD_NUMBER: builtins.int
    SCHD_TO_FIELD_NUMBER: builtins.int
    FLIGHT_FIELD_NUMBER: builtins.int
    OPERATOR_FIELD_NUMBER: builtins.int
    AC_TYPE_FIELD_NUMBER: builtins.int
    LAT_FIELD_NUMBER: builtins.int
    LON_FIELD_NUMBER: builtins.int
    CALLSIGN_FIELD_NUMBER: builtins.int
    ID_FIELD_NUMBER: builtins.int
    RESTRICTED_FIELD_NUMBER: builtins.int
    reg: builtins.str
    schd_from: builtins.str
    schd_to: builtins.str
    flight: builtins.str
    operator: builtins.str
    ac_type: builtins.str
    lat: builtins.float
    lon: builtins.float
    callsign: builtins.str
    id: builtins.int
    restricted: builtins.bool
    def __init__(
        self,
        *,
        reg: builtins.str = ...,
        schd_from: builtins.str = ...,
        schd_to: builtins.str = ...,
        flight: builtins.str = ...,
        operator: builtins.str = ...,
        ac_type: builtins.str = ...,
        lat: builtins.float = ...,
        lon: builtins.float = ...,
        callsign: builtins.str = ...,
        id: builtins.int = ...,
        restricted: builtins.bool = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["ac_type", b"ac_type", "callsign", b"callsign", "flight", b"flight", "id", b"id", "lat", b"lat", "lon", b"lon", "operator", b"operator", "reg", b"reg", "restricted", b"restricted", "schd_from", b"schd_from", "schd_to", b"schd_to"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...

@typing.final
class FetchSearchIndexRequest(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

@typing.final
class FetchSearchIndexResponse(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    FLIGHTS_LIST_FIELD_NUMBER: builtins.int
    @builtins.property
    def flights_list(self) -> containers.RepeatedCompositeFieldContainer[FlightSearchData]: ...
    def __init__(
        self,
        *,
        flights_list: abc.Iterable[FlightSearchData] | None = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["flights_list", b"flights_list"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...



# NOTE: this file is created to prevent circular imports in `authentication`
from pathlib import Path

from platformdirs import user_config_dir

PATH_CONFIG = Path(user_config_dir("fr24"))

FP_CONFIG_FILE = PATH_CONFIG / "fr24.conf"



# ruff: noqa
# mypy: disable-error-code="top-level-await, no-redef"
# %%
# --8<-- [start:script0]
import httpx
from fr24.grpc import historic_trail
from fr24.proto.v1_pb2 import HistoricTrailRequest, HistoricTrailResponse
from fr24.proto import parse_data
from fr24.proto.headers import get_grpc_headers


async def historic_trail_data() -> HistoricTrailResponse:
    headers = httpx.Headers(get_grpc_headers(auth=None))
    async with httpx.AsyncClient() as client:
        message = HistoricTrailRequest(flight_id=0x395C43CF)
        response = await historic_trail(client, message, headers)
        return parse_data(response.content, HistoricTrailResponse).unwrap()


data = await historic_trail_data()
data
# --8<-- [end:script0]
# %%
"""
# --8<-- [start:output0]
504 GATEWAY TIMEOUT
# --8<-- [end:output0]
"""



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from collections import abc as abc
from fr24.proto import _common_pb2 as _common_pb2
from google.protobuf import descriptor as descriptor
from google.protobuf import message as message
from google.protobuf.internal import containers as containers
import builtins as builtins
import sys
import typing as typing

if sys.version_info >= (3, 10):
    from typing import TypeAlias as _TypeAlias
else:
    from typing_extensions import TypeAlias as _TypeAlias

DESCRIPTOR: descriptor.FileDescriptor

@typing.final
class FollowFlightRequest(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    FLIGHT_ID_FIELD_NUMBER: builtins.int
    RESTRICTION_MODE_FIELD_NUMBER: builtins.int
    flight_id: builtins.int
    """FR24 Flight ID (e.g. `962788562` = `0x3962fcd2`)"""
    restriction_mode: _common_pb2.RestrictionVisibility.ValueType
    def __init__(
        self,
        *,
        flight_id: builtins.int = ...,
        restriction_mode: _common_pb2.RestrictionVisibility.ValueType = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["flight_id", b"flight_id", "restriction_mode", b"restriction_mode"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...

@typing.final
class FollowFlightResponse(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    AIRCRAFT_INFO_FIELD_NUMBER: builtins.int
    FLIGHT_PLAN_FIELD_NUMBER: builtins.int
    SCHEDULE_INFO_FIELD_NUMBER: builtins.int
    FLIGHT_PROGRESS_FIELD_NUMBER: builtins.int
    FLIGHT_INFO_FIELD_NUMBER: builtins.int
    FLIGHT_TRAIL_LIST_FIELD_NUMBER: builtins.int
    POSITION_BUFFER_FIELD_NUMBER: builtins.int
    @builtins.property
    def aircraft_info(self) -> _common_pb2.AircraftInfo: ...
    @builtins.property
    def flight_plan(self) -> _common_pb2.FlightPlan: ...
    @builtins.property
    def schedule_info(self) -> _common_pb2.ScheduleInfo: ...
    @builtins.property
    def flight_progress(self) -> _common_pb2.FlightProgress: ...
    @builtins.property
    def flight_info(self) -> _common_pb2.ExtendedFlightInfo: ...
    @builtins.property
    def flight_trail_list(self) -> containers.RepeatedCompositeFieldContainer[_common_pb2.TrailPoint]: ...
    @builtins.property
    def position_buffer(self) -> _common_pb2.PositionBuffer: ...
    def __init__(
        self,
        *,
        aircraft_info: _common_pb2.AircraftInfo | None = ...,
        flight_plan: _common_pb2.FlightPlan | None = ...,
        schedule_info: _common_pb2.ScheduleInfo | None = ...,
        flight_progress: _common_pb2.FlightProgress | None = ...,
        flight_info: _common_pb2.ExtendedFlightInfo | None = ...,
        flight_trail_list: abc.Iterable[_common_pb2.TrailPoint] | None = ...,
        position_buffer: _common_pb2.PositionBuffer | None = ...,
    ) -> None: ...
    _HasFieldArgType: _TypeAlias = typing.Literal["aircraft_info", b"aircraft_info", "flight_info", b"flight_info", "flight_plan", b"flight_plan", "flight_progress", b"flight_progress", "position_buffer", b"position_buffer", "schedule_info", b"schedule_info"]  # noqa: Y015
    def HasField(self, field_name: _HasFieldArgType) -> builtins.bool: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["aircraft_info", b"aircraft_info", "flight_info", b"flight_info", "flight_plan", b"flight_plan", "flight_progress", b"flight_progress", "flight_trail_list", b"flight_trail_list", "position_buffer", b"position_buffer", "schedule_info", b"schedule_info"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
NOTE: the definition in the web client is:
```
message LiveFlightsStatusResponse {
    repeated LiveFlightStatusData flights_map = 1;
}
```
but in reality it returns a wrapped `LiveFlightStatusData`.
the names `LiveFlightStatus`, `flight_id` and `data`
are non-official.
"""

from collections import abc as abc
from fr24.proto import _common_pb2 as _common_pb2
from google.protobuf import descriptor as descriptor
from google.protobuf import message as message
from google.protobuf.internal import containers as containers
import builtins as builtins
import sys
import typing as typing

if sys.version_info >= (3, 10):
    from typing import TypeAlias as _TypeAlias
else:
    from typing_extensions import TypeAlias as _TypeAlias

DESCRIPTOR: descriptor.FileDescriptor

@typing.final
class LiveFlightStatusData(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    LAT_FIELD_NUMBER: builtins.int
    LON_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    SQUAWK_FIELD_NUMBER: builtins.int
    lat: builtins.float
    lon: builtins.float
    status: _common_pb2.Status.ValueType
    squawk: builtins.int
    def __init__(
        self,
        *,
        lat: builtins.float = ...,
        lon: builtins.float = ...,
        status: _common_pb2.Status.ValueType = ...,
        squawk: builtins.int = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["lat", b"lat", "lon", b"lon", "squawk", b"squawk", "status", b"status"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...

@typing.final
class LiveFlightsStatusRequest(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    FLIGHT_IDS_LIST_FIELD_NUMBER: builtins.int
    @builtins.property
    def flight_ids_list(self) -> containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        flight_ids_list: abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["flight_ids_list", b"flight_ids_list"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...

@typing.final
class LiveFlightStatus(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    FLIGHT_ID_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    flight_id: builtins.int
    @builtins.property
    def data(self) -> LiveFlightStatusData: ...
    def __init__(
        self,
        *,
        flight_id: builtins.int = ...,
        data: LiveFlightStatusData | None = ...,
    ) -> None: ...
    _HasFieldArgType: _TypeAlias = typing.Literal["data", b"data"]  # noqa: Y015
    def HasField(self, field_name: _HasFieldArgType) -> builtins.bool: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["data", b"data", "flight_id", b"flight_id"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...

@typing.final
class LiveFlightsStatusResponse(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    FLIGHTS_MAP_FIELD_NUMBER: builtins.int
    @builtins.property
    def flights_map(self) -> containers.RepeatedCompositeFieldContainer[LiveFlightStatus]: ...
    def __init__(
        self,
        *,
        flights_map: abc.Iterable[LiveFlightStatus] | None = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["flights_map", b"flights_map"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...
